from django.conf.urls import url
from django.contrib.auth.views import LoginView, LogoutView

from . import views

urlpatterns = (
    url(
        r'^login$',
        LoginView.as_view(template_name="accounts/login.html"),
        name='accounts.login',
    ),
    url(
        r'^logout$',
        LogoutView.as_view(template_name="accounts/logout.html"),
        name='accounts.logout',
    ),
    url(
        r'^search/groups$',
        views.search_groups,
        name='accounts.search-groups',
    ),
    url(
        r'^search/users$',
        views.search_users,
        name='accounts.search-users',
    ),
    url(
        r'^users$',
        views.users,
        name='accounts.user-list',
    ),
    url(
        r'^users/(?P<uid>\w+)$',
        views.user_detail,
        name='accounts.user-detail',
    ),
    url(
        r'^users/(?P<uid>\w+)/reactivate$',
        views.user_activate,
        name='accounts.user-reactivate',
    ),
    url(
        r'^users/(?P<uid>\w+)/deactivate$',
        views.user_activate,
        {'deactivate': True},
        name='accounts.user-deactivate',
    ),
)



from django.conf import settings
from django.db import models


class LogEntry(models.Model):
    actor = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        models.PROTECT,
        blank=True,
        null=True,
        related_name='logged_actions',
    )
    group = models.ForeignKey(
        'auth.Group',
        models.PROTECT,
        blank=True,
        null=True,
        related_name='logged_actions',
    )
    message = models.TextField()
    secret = models.ForeignKey(
        'secrets.Secret',
        models.PROTECT,
        blank=True,
        null=True,
        related_name='logged_actions',
    )
    secret_revision = models.ForeignKey(
        'secrets.SecretRevision',
        models.PROTECT,
        blank=True,
        null=True,
        related_name='logged_actions',
    )
    time = models.DateTimeField(
        auto_now_add=True,
    )
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        models.PROTECT,
        blank=True,
        null=True,
        related_name='affected_by_actions',
    )

    class Meta:
        ordering = ('-time',)



from django.db import models
from django.utils.translation import ugettext_lazy as _


class Setting(models.Model):
    key = models.CharField(
        max_length=64,
        unique=True,
    )
    value = models.CharField(
        max_length=255,
    )

    class Meta:
        ordering = ('key',)

    @classmethod
    def get(cls, key, **kwargs):
        try:
            return cls.objects.get(key=key).value
        except cls.DoesNotExist:
            try:
                return kwargs['default']
            except KeyError:
                raise KeyError(_("value for '{}' not set").format(key))

    @classmethod
    def set(cls, key, value):
        try:
            setting = cls.objects.get(key=key)
        except cls.DoesNotExist:
            setting = cls()
            setting.key = key
        setting.value = value
        setting.save()



from django.core.management.base import BaseCommand, CommandError

from ...models import Secret


class Command(BaseCommand):
    help = 'Update search index'

    def handle(self, *args, **options):
        secrets_total = Secret.objects.count()
        secrets_processed = 0
        last_progress = 0
        for secret in Secret.objects.all():
            secret.save()
            secrets_processed += 1
            progress = int((secrets_processed / secrets_total) * 100)
            if progress > last_progress:
                self.stdout.write("{}%".format(progress))
                last_progress = progress

        self.stdout.write(self.style.SUCCESS(
            "Finished updating search index for {} objects.".format(secrets_total)
        ))



# Generated by Django 2.0 on 2018-02-20 10:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('secrets', '0014_auto_20170313_1544'),
    ]

    operations = [
        migrations.AddField(
            model_name='secretrevision',
            name='plaintext_data_sha256',
            field=models.CharField(default='<unknown>', max_length=64),
            preserve_default=False,
        ),
    ]



from django.contrib import admin
from django.utils.translation import ugettext_lazy as _

from .models import LogEntry


class LogEntryAdmin(admin.ModelAdmin):
    fieldsets = (
        (None, {
            'fields': (
                'time',
                'message',
            ),
        }),
        (_("Linked objects"), {
            'fields': (
                'actor',
                'secret',
                'secret_revision',
                'group',
                'user',
            ),
        }),
    )
    date_hierarchy = 'time'
    list_display = ('time', 'message')
    readonly_fields = (
        'time',
        'message',
        'actor',
        'secret',
        'secret_revision',
        'group',
        'user',
    )
    search_fields = ('message', 'actor__username')

admin.site.register(LogEntry, LogEntryAdmin)



from django.contrib import admin
from django.utils.translation import ugettext_lazy as _

from .models import AccessRequest, Secret, SecretRevision


class AccessRequestAdmin(admin.ModelAdmin):
    fieldsets = (
        (_("Subject"), {
            'fields': (
                'requester',
                'secret',
                'created',
                'reason_request',
            ),
        }),
        (_("Status"), {
            'fields': (
                'reviewers',
                'status',
                'closed',
                'closed_by',
                'reason_rejected',
            ),
        }),
    )
    date_hierarchy = 'created'
    list_display = ('requester', 'secret', 'status', 'created')
    list_filter = ('status',)
    readonly_fields = ('created',)
    search_fields = ('requester__username', 'password__name',)

admin.site.register(AccessRequest, AccessRequestAdmin)


class SecretAdmin(admin.ModelAdmin):
    fieldsets = (
        (None, {
            'fields': (
                'name',
                'content_type',
                'description',
            ),
        }),
        (_("Audit"), {
            'fields': (
                'created',
                'last_read',
            ),
        }),
        (_("Security"), {
            'fields': (
                'status',
                'access_policy',
                'needs_changing_on_leave',
                'allowed_groups',
                'allowed_users',
            ),
        }),
    )
    date_hierarchy = 'created'
    list_display = ('name', 'last_read')
    list_filter = ('access_policy', 'needs_changing_on_leave', 'status')
    radio_fields = {
        'access_policy': admin.HORIZONTAL,
        'status': admin.HORIZONTAL,
    }
    readonly_fields = ('created', 'last_read')
    search_fields = ('name', 'description')

admin.site.register(Secret, SecretAdmin)


class SecretRevisionAdmin(admin.ModelAdmin):
    fieldsets = (
        (None, {
            'fields': (
                'secret',
                'encrypted_data',
            ),
        }),
        (_("Audit"), {
            'fields': (
                'created',
                'set_by',
                'accessed_by',
            ),
        }),
    )
    date_hierarchy = 'created'
    list_display = ('secret', 'id', 'created')
    readonly_fields = ('accessed_by', 'created', 'set_by')
    search_fields = ('secret__name',)

admin.site.register(SecretRevision, SecretRevisionAdmin)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('secrets', '0007_auto_20150205_1918'),
    ]

    operations = [
        migrations.AddField(
            model_name='accessrequest',
            name='hashid',
            field=models.CharField(unique=True, max_length=24, null=True),
        ),
        migrations.AddField(
            model_name='secret',
            name='hashid',
            field=models.CharField(unique=True, max_length=24, null=True),
        ),
        migrations.AddField(
            model_name='secretrevision',
            name='hashid',
            field=models.CharField(unique=True, max_length=24, null=True),
        ),
    ]



from base64 import decodestring, encodestring
from configparser import SafeConfigParser
from gettext import gettext as _
from hashlib import sha1
from os import environ, umask
from os.path import exists, isfile
from random import choice
from string import ascii_letters, digits, punctuation
from urllib.parse import urlparse

from cryptography.fernet import Fernet


def configure_base_url(config, settings):
    settings.BASE_URL = config.get("teamvault", "base_url")
    settings.ALLOWED_HOSTS = [urlparse(settings.BASE_URL).hostname]


def configure_database(config):
    """
    Called directly from the Django settings module.
    """
    DATABASES = {
        'default': {
            'ENGINE': 'django.db.backends.postgresql_psycopg2',
            'HOST': get_from_config(config, "database", "host", "localhost"),
            'NAME': get_from_config(config, "database", "name", "teamvault"),
            'PASSWORD': get_from_config(config, "database", "password", ""),
            'PORT': get_from_config(config, "database", "port", "5432"),
            'USER': get_from_config(config, "database", "user", "teamvault"),
        },
    }
    return DATABASES


def configure_debugging(config, settings):
    enabled = get_from_config(config, "teamvault", "insecure_debug_mode", "no")
    if enabled.lower() in ("1", "enabled", "true", "yes"):
        settings.DEBUG = True
        settings.EMAIL_BACKEND = 'django.core.mail.backends.console.EmailBackend'
    else:
        settings.DEBUG = False


def configure_django_secret_key(config):
    """
    Called directly from the Django settings module.
    """
    return decodestring(config.get("django", "secret_key").encode()).decode('utf-8')


def configure_hashid(config):
    """
    Called directly from the Django settings module.
    """
    return (
        int(get_from_config(config, "hashid", "min_length", "6")),
        decodestring(config.get("hashid", "salt").encode()).decode('utf-8'),
    )


def configure_ldap_auth(config, settings):
    if not config.has_section("auth_ldap"):
        return

    from django_auth_ldap.config import LDAPSearch, MemberDNGroupType
    from ldap import SCOPE_SUBTREE

    settings.AUTHENTICATION_BACKENDS = (
        'django_auth_ldap.backend.LDAPBackend',
        'django.contrib.auth.backends.ModelBackend',
    )
    settings.AUTH_LDAP_SERVER_URI = config.get("auth_ldap", "server_uri")
    settings.AUTH_LDAP_BIND_DN = config.get("auth_ldap", "bind_dn")
    settings.AUTH_LDAP_BIND_PASSWORD = config.get("auth_ldap", "password")

    settings.AUTH_LDAP_USER_SEARCH = LDAPSearch(
        config.get("auth_ldap", "user_base_dn"),
        SCOPE_SUBTREE,
        get_from_config(config, "auth_ldap", "user_search_filter", "(cn=%(user)s)"),
    )
    settings.AUTH_LDAP_GROUP_SEARCH = LDAPSearch(
        config.get("auth_ldap", "group_base_dn"),
        SCOPE_SUBTREE,
        get_from_config(config, "auth_ldap", "group_search_filter", "(objectClass=group)"),
    )

    settings.AUTH_LDAP_GROUP_TYPE = MemberDNGroupType('member')
    settings.AUTH_LDAP_REQUIRE_GROUP = get_from_config(config, "auth_ldap", "require_group", None)

    settings.AUTH_LDAP_USER_ATTR_MAP = {
        "email": get_from_config(config, "auth_ldap", "attr_email", "mail"),
        "first_name": get_from_config(config, "auth_ldap", "attr_first_name", "givenName"),
        "last_name": get_from_config(config, "auth_ldap", "attr_last_name", "sn"),
    }
    settings.AUTH_LDAP_USER_FLAGS_BY_GROUP = {
        "is_staff": config.get("auth_ldap", "admin_group"),
        "is_superuser": config.get("auth_ldap", "admin_group"),
    }

    settings.AUTH_LDAP_ALWAYS_UPDATE_USER = True
    settings.AUTH_LDAP_FIND_GROUP_PERMS = False
    settings.AUTH_LDAP_MIRROR_GROUPS = True
    settings.AUTH_LDAP_CACHE_GROUPS = True
    settings.AUTH_LDAP_GROUP_CACHE_TIMEOUT = 900


def configure_logging(config):
    level = 'INFO'

    if get_from_config(config, "teamvault", "insecure_debug_mode", "no").lower() in \
        ("1", "enabled", "true", "yes"):
        level = 'DEBUG'

    LOGGING = {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'console': {
                'format': "[%(asctime)s] %(levelname)s %(module)s: %(message)s",
            },
        },
        'handlers': {
            'console': {
                'formatter': 'console',
                'level': 'DEBUG',
                'class': 'logging.StreamHandler',
            },
        },
        'loggers': {
            'django': {
                'handlers': ['console'],
                'level': level,
            },
            'teamvault': {
                'handlers': ['console'],
                'level': level,
            },
        },
    }

    return LOGGING


def configure_max_file_size(config, settings):
    settings.TEAMVAULT_MAX_FILE_SIZE = int(
        get_from_config(config, "teamvault", "max_file_size", "5242880")
    )
    settings.FILE_UPLOAD_MAX_MEMORY_SIZE = settings.TEAMVAULT_MAX_FILE_SIZE


def configure_session(config):
    """
    Called directly from the Django settings module.
    """
    age = int(get_from_config(config, "teamvault", "session_cookie_age", "3600"))
    expire = get_from_config(config, "teamvault", "session_expire_at_browser_close", "True")
    secure = get_from_config(config, "teamvault", "session_cookie_secure", "False")

    if age <= 0:
        age = 3600
    expire = expire.lower() in ("1", "enabled", "true", "yes")
    secure = secure.lower() in ("1", "enabled", "true", "yes")

    return age, expire, secure


def configure_teamvault_secret_key(config, settings):
    from .models import Setting

    checksum = Setting.get("fernet_key_hash", default=None)
    key = config.get("teamvault", "fernet_key")
    key_hash = sha1(key.encode('utf-8')).hexdigest()

    if checksum is None:
        Setting.set("fernet_key_hash", key_hash)

    elif key_hash != checksum:
        raise RuntimeError(_(
            "secret in '{path}' does not match SHA1 hash in database ({hash})"
        ).format(
            hash=checksum,
            path=environ['TEAMVAULT_CONFIG_FILE'],
        ))

    settings.TEAMVAULT_SECRET_KEY = key


def create_default_config(filename):
    if exists(filename):
        raise RuntimeError("not overwriting existing path {}".format(filename))
    SECRET_KEY = "".join(choice(ascii_letters + digits + punctuation) for i in range(50))
    HASHID_SALT = "".join(choice(ascii_letters + digits + punctuation) for i in range(50))
    config = """
[teamvault]
# Set this to the URL users use to reach the application
base_url = https://example.com
# This key has been generated for you, there is no need to change it
fernet_key = {teamvault_key}
# do not enable this in production
insecure_debug_mode = disabled
# file uploads larger than this number of bytes will have their connection reset
max_file_size = 5242880
session_cookie_age = 3600
session_expire_at_browser_close = True
session_cookie_secure = False

[django]
# This key has been generated for you, there is no need to change it
secret_key = {django_key}

[database]
host = localhost
name = teamvault
user = teamvault
password = teamvault

[hashid]
min_length = 6
# This salt has been generated for you, there is no need to change it
salt = {hashid_salt}

#[auth_ldap]
#server_uri = ldaps://ldap.example.com
#bind_dn = cn=root,dc=example,dc=com
#password = example
#user_base_dn = ou=users,dc=example,dc=com
##user_search_filter = (cn=%%(user)s)
#group_base_dn = ou=groups,dc=example,dc=com
##group_search_filter = (objectClass=group)
##require_group = cn=employees,ou=groups,dc=example,dc=com
##attr_email = mail
##attr_first_name = givenName
##attr_last_name = sn
#admin_group = cn=admins,ou=groups,dc=example,dc=com
    """.format(
        django_key=encodestring(SECRET_KEY.encode('utf-8')).decode('utf-8'),
        hashid_salt=encodestring(HASHID_SALT.encode('utf-8')).decode('utf-8'),
        teamvault_key=Fernet.generate_key().decode('utf-8'),
    )
    old_umask = umask(7)
    try:
        with open(filename, 'wt') as f:
            f.write(config.strip())
    finally:
        umask(old_umask)


def get_config():
    if not isfile(environ['TEAMVAULT_CONFIG_FILE']):
        raise RuntimeError(
            _("missing config file at {} "
              "(set env var TEAMVAULT_CONFIG_FILE to use a different path)").format(
                environ['TEAMVAULT_CONFIG_FILE'],
            )
        )

    with open(environ['TEAMVAULT_CONFIG_FILE']) as f:
        # SafeConfigParser.read() will not complain if it can't read the
        # file, so we need to read it once ourselves to get a proper IOError
        f.read()

    config = SafeConfigParser()
    config.read(environ['TEAMVAULT_CONFIG_FILE'])
    return config


def get_from_config(config, section, option, default):
    if config.has_option(section, option):
        return config.get(section, option)
    else:
        return default



# Generated by Django 2.0 on 2018-02-20 14:28

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('secrets', '0020_auto_20180220_1356'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='secret',
            name='owner_groups',
        ),
        migrations.RemoveField(
            model_name='secret',
            name='owner_users',
        ),
    ]



from json import dumps

from django.contrib.auth.decorators import login_required
from django.contrib.auth.decorators import user_passes_test
from django.contrib.auth.models import User, Group
from django.db import transaction
from django.http import HttpResponse, HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _
from django.views.decorators.http import require_http_methods
from django.views.generic import DetailView, ListView

from ..audit.auditlog import log
from ..secrets.models import Secret, SecretRevision


@login_required
def search_groups(request):
    search_term = request.GET['q']
    search_result = {'results': []}
    groups = Group.objects.filter(name__icontains=search_term)
    for group in groups:
        search_result['results'].append({'id': group.id, 'text': group.name})
    return HttpResponse(dumps(search_result), content_type="application/json")


@login_required
def search_users(request):
    search_term = request.GET['q']
    search_result = {'results': []}
    users = User.objects.filter(
        is_active=True,
        username__icontains=search_term,
    )
    for user in users:
        search_result['results'].append({'id': user.id, 'text': user.username})
    return HttpResponse(dumps(search_result), content_type="application/json")


class UserList(ListView):
    context_object_name = 'users'
    paginate_by = 100
    template_name = "accounts/user_list.html"

    def get_queryset(self):
        return User.objects.order_by('username')
users = user_passes_test(lambda u: u.is_superuser)(UserList.as_view())


class UserDetail(DetailView):
    context_object_name = 'user'
    template_name = "accounts/user_detail.html"

    def get_object(self):
        return get_object_or_404(
            User,
            id=self.kwargs['uid'],
        )
user_detail = user_passes_test(lambda u: u.is_superuser)(UserDetail.as_view())


@user_passes_test(lambda u: u.is_superuser)
@require_http_methods(["POST"])
def user_activate(request, uid, deactivate=False):
    user = get_object_or_404(
        User,
        id=uid,
        is_active=deactivate,
    )
    user.is_active = not deactivate
    user.save()
    if deactivate:
        accessed_revs = SecretRevision.objects.filter(
            accessed_by=user,
        ).exclude(
            secret__needs_changing_on_leave=False,
        ).exclude(
            secret__status=Secret.STATUS_NEEDS_CHANGING,
        ).select_related(
            'secret',
        )
        secrets = set()
        for rev in accessed_revs:
            if rev.is_current_revision:
                secrets.add(rev.secret)
        with transaction.atomic():
            for secret in secrets:
                secret.status = Secret.STATUS_NEEDS_CHANGING
                secret.save()
                msg = _(
                    "secret '{secret}' needs changing because user '{user}' was deactivated"
                ).format(
                    secret=secret.name,
                    user=user.username,
                )
                log(msg, actor=request.user, secret=secret, user=user)
        log(
            _("{actor} deactivated {user}, {secrets} secrets marked for changing").format(
                actor=request.user.username,
                user=user.username,
                secrets=len(secrets),
            ),
            actor=request.user,
            user=user,
        )
    else:
        log(
            _("{actor} reactivated {user}").format(
                actor=request.user.username,
                user=user.username,
            ),
            actor=request.user,
            user=user,
        )
    return HttpResponseRedirect(reverse('accounts.user-detail', kwargs={'uid': user.id}))



# -*- coding: utf-8 -*-
# Generated by Django 1.10.2 on 2016-10-21 14:11
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('auth', '0008_alter_user_username_max_length'),
        ('secrets', '0012_secret_search_index'),
    ]

    operations = [
        migrations.AddField(
            model_name='secret',
            name='owner_groups',
            field=models.ManyToManyField(blank=True, related_name='owned_passwords', to='auth.Group'),
        ),
        migrations.AddField(
            model_name='secret',
            name='owner_users',
            field=models.ManyToManyField(blank=True, related_name='owned_passwords', to=settings.AUTH_USER_MODEL),
        ),
    ]



from base64 import decodestring, encodestring
from json import dumps, loads

from django.contrib.auth.models import Group, User
from django.shortcuts import get_object_or_404
from rest_framework import generics
from rest_framework import serializers
from rest_framework import status
from rest_framework.decorators import api_view
from rest_framework.response import Response
from rest_framework.reverse import reverse

from .models import AccessRequest, Secret, SecretRevision


ACCESS_POLICY_REPR = {
    Secret.ACCESS_POLICY_ANY: "any",
    Secret.ACCESS_POLICY_HIDDEN: "hidden",
    Secret.ACCESS_POLICY_REQUEST: "request",
}
REPR_ACCESS_POLICY = {v: k for k, v in ACCESS_POLICY_REPR.items()}

ACCESS_REQUEST_STATUS_REPR = {
    AccessRequest.STATUS_PENDING: "pending",
    AccessRequest.STATUS_REJECTED: "rejected",
    AccessRequest.STATUS_APPROVED: "approved",
}
ACCESS_REQUEST_REPR_STATUS = {v: k for k, v in ACCESS_REQUEST_STATUS_REPR.items()}

CONTENT_TYPE_REPR = {
    Secret.CONTENT_CC: "cc",
    Secret.CONTENT_FILE: "file",
    Secret.CONTENT_PASSWORD: "password",
}
REPR_CONTENT_TYPE = {v: k for k, v in CONTENT_TYPE_REPR.items()}

SECRET_STATUS_REPR = {
    Secret.STATUS_DELETED: "deleted",
    Secret.STATUS_NEEDS_CHANGING: "needs_changing",
    Secret.STATUS_OK: "ok",
}
SECRET_REPR_STATUS = {v: k for k, v in SECRET_STATUS_REPR.items()}

REQUIRED_CC_FIELDS = set((
    'holder', 'expiration_month', 'expiration_year', 'number', 'security_code',
))


def _extract_data(validated_data):
    if 'password' in validated_data and not \
            REQUIRED_CC_FIELDS.intersection(set(validated_data.keys())):
        return (
            validated_data.pop('password'),
            Secret.CONTENT_PASSWORD,
        )
    elif 'file' in validated_data:
        return (
            decodestring(validated_data.pop('file').encode('ascii')),
            Secret.CONTENT_FILE,
        )
    elif REQUIRED_CC_FIELDS.intersection(set(validated_data.keys())):
        data = {}
        for cc_field in REQUIRED_CC_FIELDS:
            try:
                data[cc_field] = str(validated_data.pop(cc_field))
            except KeyError:
                raise serializers.ValidationError(
                    "missing required CC field '{}'".format(cc_field)
                )
        if 'password' in validated_data:
            data['password'] = validated_data.pop('password')
        else:
            data['password'] = ""
        return (
            dumps(data),
            Secret.CONTENT_CC,
        )
    else:
        return (None, None)


class AccessRequestSerializer(serializers.HyperlinkedModelSerializer):
    api_url = serializers.HyperlinkedIdentityField(
        lookup_field='hashid',
        view_name='api.access-request_detail',
    )
    closed_by = serializers.SlugRelatedField(
        default=serializers.CurrentUserDefault(),
        read_only=True,
        slug_field='username',
    )
    requester = serializers.SlugRelatedField(
        default=serializers.CurrentUserDefault(),
        read_only=True,
        slug_field='username',
    )
    secret = serializers.HyperlinkedRelatedField(
        lookup_field='hashid',
        queryset=Secret.objects.exclude(status=Secret.STATUS_DELETED),
        view_name='api.secret_detail',
    )
    reviewers = serializers.SlugRelatedField(
        many=True,
        queryset=User.objects.exclude(is_active=False),
        slug_field='username',
    )
    web_url = serializers.CharField(
        required=False,
        read_only=True,
    )

    def to_representation(self, instance):
        rep = super(AccessRequestSerializer, self).to_representation(instance)
        rep['web_url'] = instance.full_url
        rep['status'] = ACCESS_REQUEST_STATUS_REPR[rep['status']]
        return rep

    class Meta:
        model = AccessRequest
        fields = (
            'api_url',
            'closed',
            'closed_by',
            'created',
            'secret',
            'reason_request',
            'reason_rejected',
            'requester',
            'reviewers',
            'status',
            'web_url',
        )
        read_only_fields = (
            'closed',
            'closed_by',
            'created',
        )


class AccessRequestDetail(generics.RetrieveUpdateAPIView):
    model = AccessRequest
    serializer_class = AccessRequestSerializer

    def get_object(self):
        obj = get_object_or_404(AccessRequest, hashid=self.kwargs['hashid'])
        if (
            self.request.user != obj.requester and
            self.request.user not in obj.reviewers and
            not self.request.user.is_superuser
        ):
            self.permission_denied(self.request)
        return obj

    def pre_save(self, obj):
        previous_state = AccessRequest.objects.get(hashid=obj.hashid)
        if (
            previous_state.status != AccessRequest.STATUS_PENDING or
            obj.status == AccessRequest.STATUS_PENDING or
            self.request.user not in previous_state.reviewers.all() or
            obj.password != previous_state.password
        ):
            self.permission_denied(self.request)

    def post_save(self, obj, created=False):
        #obj.handle_status() TODO
        pass


class AccessRequestList(generics.ListCreateAPIView):
    model = AccessRequest
    serializer_class = AccessRequestSerializer

    def get_queryset(self):
        return AccessRequest.get_all_readable_by_user(self.request.user)

    def pre_save(self, obj):
        obj.reason_rejected = ""
        obj.requester = self.request.user
        obj.status = AccessRequest.STATUS_PENDING

    def post_save(self, obj, created=False):
        if created:
            obj.assign_reviewers()


class SecretRevisionSerializer(serializers.HyperlinkedModelSerializer):
    api_url = serializers.HyperlinkedIdentityField(
        lookup_field='hashid',
        view_name='api.secret-revision_detail',
    )
    data_url = serializers.CharField(
        read_only=True,
        required=False,
        source='id',
    )
    set_by = serializers.SlugRelatedField(
        read_only=True,
        slug_field='username',
    )

    def to_representation(self, instance):
        rep = super(SecretRevisionSerializer, self).to_representation(instance)
        rep['data_url'] = reverse(
            'api.secret-revision_data',
            kwargs={'hashid': instance.hashid},
            request=self.context['request'],
        )
        return rep

    class Meta:
        model = SecretRevision
        fields = (
            'api_url',
            'created',
            'data_url',
            'set_by',
        )
        read_only_fields = (
            'created',
        )


class SecretSerializer(serializers.HyperlinkedModelSerializer):
    allowed_groups = serializers.SlugRelatedField(
        many=True,
        queryset=Group.objects.all(),
        required=False,
        slug_field='name',
    )
    allowed_users = serializers.SlugRelatedField(
        many=True,
        queryset=User.objects.exclude(is_active=False),
        required=False,
        slug_field='username',
    )
    api_url = serializers.HyperlinkedIdentityField(
        lookup_field='hashid',
        view_name='api.secret_detail',
    )
    created_by = serializers.SlugRelatedField(
        default=serializers.CurrentUserDefault(),
        read_only=True,
        slug_field='username',
    )
    current_revision = serializers.HyperlinkedRelatedField(
        lookup_field='hashid',
        read_only=True,
        view_name='api.secret-revision_detail',
    )
    data_readable = serializers.BooleanField(
        read_only=True,
        required=False,
        source='id',
    )
    expiration_month = serializers.CharField(
        required=False,
        write_only=True,
    )
    expiration_year = serializers.CharField(
        required=False,
        write_only=True,
    )
    file = serializers.CharField(
        required=False,
        write_only=True,
    )
    holder = serializers.CharField(
        required=False,
        write_only=True,
    )
    notify_on_access_request = serializers.SlugRelatedField(
        many=True,
        queryset=User.objects.exclude(is_active=False),
        required=False,
        slug_field='username',
    )
    number = serializers.CharField(
        required=False,
        write_only=True,
    )
    password = serializers.CharField(
        required=False,
        write_only=True,
    )
    security_code = serializers.CharField(
        required=False,
        write_only=True,
    )
    web_url = serializers.CharField(
        required=False,
        read_only=True,
    )

    def create(self, validated_data):
        allowed_groups = validated_data.pop('allowed_groups', [])
        allowed_users = validated_data.pop('allowed_users', [])
        notify_on_access_request = validated_data.pop('notify_on_access_request', [])

        data, content_type = _extract_data(validated_data)
        if not data:
            raise serializers.ValidationError("missing secret field (e.g. 'password')")

        instance = self.Meta.model.objects.create(**validated_data)

        instance.allowed_groups = allowed_groups
        instance.allowed_users = allowed_users
        instance.content_type = content_type
        instance.notify_on_access_request = notify_on_access_request
        instance._data = data
        return instance

    def to_internal_value(self, data):
        try:
            data['access_policy'] = REPR_ACCESS_POLICY[data.get('access_policy', None)]
        except KeyError:
            # Validation will catch it
            pass

        try:
            data['status'] = SECRET_REPR_STATUS[data.get('status', None)]
        except KeyError:
            # Validation will catch it
            pass

        return super(SecretSerializer, self).to_internal_value(data)

    def to_representation(self, instance):
        rep = super(SecretSerializer, self).to_representation(instance)
        rep['access_policy'] = ACCESS_POLICY_REPR[rep['access_policy']]
        rep['content_type'] = CONTENT_TYPE_REPR[rep['content_type']]
        rep['data_readable'] = instance.is_readable_by_user(self.context['request'].user)
        rep['status'] = SECRET_STATUS_REPR[rep['status']]
        rep['web_url'] = instance.full_url
        return rep

    def update(self, instance, validated_data):
        data, content_type = _extract_data(validated_data)
        if content_type and content_type != instance.content_type:
            raise serializers.ValidationError("wrong secret content type")
        if data:
            instance._data = data
        return instance

    def validate(self, data):
        if 'file' in data or 'filename' in data:
            if not ('file' in data and 'filename' in data):
                raise serializers.ValidationError("must include both file and filename")
        return data

    class Meta:
        model = Secret
        fields = (
            'access_policy',
            'allowed_groups',
            'allowed_users',
            'api_url',
            'content_type',
            'created',
            'created_by',
            'current_revision',
            'data_readable',
            'description',
            'expiration_month',
            'expiration_year',
            'file',
            'filename',
            'holder',
            'last_read',
            'name',
            'needs_changing_on_leave',
            'notify_on_access_request',
            'number',
            'password',
            'security_code',
            'status',
            'url',
            'username',
            'web_url',
        )
        read_only_fields = (
            'content_type',
            'created',
            'last_read',
        )


class SecretDetail(generics.RetrieveUpdateDestroyAPIView):
    model = Secret
    serializer_class = SecretSerializer

    def destroy(self, request, *args, **kwargs):
        obj = self.get_object()
        obj.status = Secret.STATUS_DELETED
        obj.save()
        return Response(status=status.HTTP_204_NO_CONTENT)

    def get_object(self):
        obj = get_object_or_404(Secret, hashid=self.kwargs['hashid'])
        if not obj.is_visible_to_user(self.request.user):
            self.permission_denied(self.request)
        return obj

    def perform_update(self, serializer):
        instance = serializer.save()
        if hasattr(instance, '_data'):
            instance.set_data(self.request.user, instance._data)
            del instance._data


class SecretList(generics.ListCreateAPIView):
    model = Secret
    serializer_class = SecretSerializer

    def get_queryset(self):
        if 'search' in self.request.query_params:
            return Secret.get_search_results(
                self.request.user,
                self.request.query_params['search'],
            )
        else:
            return Secret.get_all_visible_to_user(self.request.user)

    def perform_create(self, serializer):
        instance = serializer.save()
        if hasattr(instance, '_data'):
            instance.set_data(self.request.user, instance._data, skip_access_check=True)
            del instance._data


class SecretRevisionDetail(generics.RetrieveAPIView):
    model = SecretRevision
    serializer_class = SecretRevisionSerializer

    def get_object(self):
        obj = get_object_or_404(SecretRevision, hashid=self.kwargs['hashid'])
        obj.secret.check_access(self.request.user)
        return obj


@api_view(['GET'])
def data_get(request, hashid):
    secret_revision = get_object_or_404(SecretRevision, hashid=hashid)
    secret_revision.secret.check_access(request.user)
    data = secret_revision.secret.get_data(request.user)
    if secret_revision.secret.content_type == Secret.CONTENT_PASSWORD:
        return Response({'password': data})
    elif secret_revision.secret.content_type == Secret.CONTENT_FILE:
        return Response({'file': encodestring(data).decode('ascii')})
    elif secret_revision.secret.content_type == Secret.CONTENT_CC:
        return Response(loads(data))



from django.shortcuts import render


def handler404(request, exception, **kwargs):
    if request.user.is_authenticated:
        return render(request, "404_loggedin.html", status=404)
    else:
        return render(request, "404_anon.html", status=404)



from django.contrib import admin

from .models import Setting


class SettingAdmin(admin.ModelAdmin):
    fieldsets = (
        (None, {
            'fields': (
                'key',
                'value',
            ),
        }),
    )
    list_display = ('key', 'value')
    search_fields = ('key',)

admin.site.register(Setting, SettingAdmin)



from datetime import date
from json import dumps

from django import forms
from django.contrib.auth.models import Group, User
from django.forms.widgets import SelectMultiple
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _

from .models import Secret


GENERIC_FIELDS_HEADER = ['name']
GENERIC_FIELDS_FOOTER = [
    'description',
    'access_policy',
    'needs_changing_on_leave',
    'notify_on_access_request',
    'allowed_groups',
    'allowed_users',
]


class Select2DataWidget(SelectMultiple):
    """
    Used to render form values as a select2-compatible data structure.
    """
    def render(self, name, value, attrs=None, choices=(), renderer=None):
        if value is None:
            return "[]"
        output = []
        value_ints = [int(v) for v in value]
        for option_id, option_label in self.choices:
            if option_id in value_ints:
                output.append({'id': str(option_id), 'text': option_label})
        return mark_safe(dumps(output))


class SecretForm(forms.ModelForm):
    allowed_groups = forms.ModelMultipleChoiceField(
        queryset=Group.objects.all(),
        required=False,
        widget=Select2DataWidget,
    )
    allowed_users = forms.ModelMultipleChoiceField(
        queryset=User.objects.filter(is_active=True),
        required=False,
        widget=Select2DataWidget,
    )
    notify_on_access_request = forms.BooleanField(
        initial=True,
        required=False,
    )


class CCForm(SecretForm):
    expiration_month = forms.IntegerField(
        min_value=1,
        max_value=12,
    )
    expiration_year = forms.IntegerField(
        min_value=date.today().year,
        max_value=date.today().year + 50,
    )
    holder = forms.CharField()
    number = forms.CharField()
    password = forms.CharField(
        required=False,
        widget=forms.PasswordInput,
    )
    security_code = forms.CharField(
        required=False,
        widget=forms.NumberInput,
    )

    class Meta:
        model = Secret
        fields = (
            GENERIC_FIELDS_HEADER +
            [
                'expiration_month',
                'expiration_year',
                'holder',
                'number',
                'password',
                'security_code',
            ] +
            GENERIC_FIELDS_FOOTER
        )


class FileForm(SecretForm):
    file = forms.FileField(
        allow_empty_file=False,
        required=False,
    )

    class Meta:
        model = Secret
        fields = (
            GENERIC_FIELDS_HEADER +
            ['file'] +
            GENERIC_FIELDS_FOOTER
        )


class PasswordForm(SecretForm):
    password = forms.CharField(
        required=False,
    )
    url = forms.CharField(
        max_length=255,
        required=False,
    )
    username = forms.CharField(
        max_length=255,
        required=False,
    )

    def clean_password(self):
        if self.instance.pk is None and not self.cleaned_data['password']:
            # password is only required when adding a new secret
            raise forms.ValidationError(_("Please enter a password."))
        return self.cleaned_data['password']

    class Meta:
        model = Secret
        fields = (
            GENERIC_FIELDS_HEADER +
            ['password', 'username', 'url'] +
            GENERIC_FIELDS_FOOTER
        )



from argparse import ArgumentParser
from gettext import gettext as _
from os import environ, mkdir
from shutil import rmtree
from subprocess import Popen
from sys import argv

from django.core.management import execute_from_command_line

from .apps.settings.config import create_default_config
from . import VERSION_STRING


def build_parser():
    parser = ArgumentParser(prog="teamvault")
    parser.add_argument(
        "--version",
        action='version',
        version=VERSION_STRING,
    )
    subparsers = parser.add_subparsers(
        title=_("subcommands"),
        help=_("use 'teamvault <subcommand> --help' for more info"),
    )

    # teamvault plumbing
    parser_plumbing = subparsers.add_parser("plumbing")
    parser_plumbing.add_argument('plumbing_command', nargs='+')
    parser_plumbing.set_defaults(func=plumbing)

    # teamvault run
    parser_run = subparsers.add_parser("run")
    parser_run.add_argument('--bind', nargs='?', help='define bind, default is 127.0.0.1:8000')
    parser_run.set_defaults(func=run)

    # teamvault setup
    parser_setup = subparsers.add_parser("setup")
    parser_setup.set_defaults(func=setup)

    # teamvault upgrade
    parser_upgrade = subparsers.add_parser("upgrade")
    parser_upgrade.set_defaults(func=upgrade)
    return parser


def main(*args):
    """
    Entry point for the 'teamvault' command line utility.

    args:   used for integration tests
    """
    if not args:
        args = argv[1:]

    parser = build_parser()
    pargs = parser.parse_args(args)
    if not hasattr(pargs, 'func'):
        parser.print_help()
        exit(2)
    pargs.func(pargs)


def plumbing(pargs):
    environ['DJANGO_SETTINGS_MODULE'] = 'teamvault.settings'
    environ.setdefault("TEAMVAULT_CONFIG_FILE", "/etc/teamvault.cfg")
    execute_from_command_line([""] + pargs.plumbing_command[0].split(" "))


def run(pargs):
    cmd = "gunicorn --preload teamvault.wsgi:application"
    if pargs.bind:
        cmd += ' -b ' + pargs.bind

    gunicorn = Popen(
        cmd,
        shell=True,
    )
    gunicorn.communicate()


def setup(pargs):
    environ.setdefault("TEAMVAULT_CONFIG_FILE", "/etc/teamvault.cfg")
    create_default_config(environ['TEAMVAULT_CONFIG_FILE'])


def upgrade(pargs):
    environ['DJANGO_SETTINGS_MODULE'] = 'teamvault.settings'
    environ.setdefault("TEAMVAULT_CONFIG_FILE", "/etc/teamvault.cfg")

    print("\n### Running migrations...\n")
    execute_from_command_line(["", "migrate", "--noinput", "-v", "3", "--traceback"])

    print("\n### Gathering static files...\n")
    from django.conf import settings
    try:
        rmtree(settings.STATIC_ROOT)
    except FileNotFoundError:
        pass
    mkdir(settings.STATIC_ROOT)
    execute_from_command_line(["", "collectstatic", "--noinput"])

    print("\n### Updating search index...\n")
    execute_from_command_line(["", "update_search_index"])



from datetime import timedelta
from hashlib import sha256
from operator import itemgetter
from random import sample

from cryptography.fernet import Fernet
from django.conf import settings
from django.contrib.auth.models import Group, User
from django.contrib.postgres.search import SearchVector, SearchVectorField
from django.core.exceptions import PermissionDenied, ValidationError
from django.db import models
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.http import Http404
from django.urls import reverse
from django.utils.timezone import now
from django.utils.translation import ugettext_lazy as _
from hashids import Hashids

from ...utils import send_mail
from ..audit.auditlog import log
from ..audit.models import LogEntry
from .exceptions import PermissionError


def validate_url(value):
    if "://" not in value or \
            value.startswith("javascript:") or \
            value.startswith("data:"):
        raise ValidationError(_("invalid URL"))


class HashIDModel(models.Model):
    hashid = models.CharField(
        max_length=24,
        null=True,
        unique=True,
    )

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        if not self.pk:
            # need to save once to get a primary key, then once again to
            # save the hashid
            super(HashIDModel, self).save(*args, **kwargs)
        if not self.hashid:
            # We cannot use the same salt for every model because
            # 1. sequentially create lots of secrets
            # 2. note the hashid of each secrets
            # 3. you can now enumerate access requests by using the same
            #    hashids
            # it's not a huge deal, but let's avoid it anyway
            hasher = Hashids(
                min_length=settings.HASHID_MIN_LENGTH,
                salt=self.HASHID_NAMESPACE + settings.HASHID_SALT,
            )
            self.hashid = hasher.encode(self.pk)
        # we cannot force insert anymore because we might already have
        # created the object
        kwargs['force_insert'] = False
        return super(HashIDModel, self).save(*args, **kwargs)


class AccessRequest(HashIDModel):
    HASHID_NAMESPACE = "AccessRequest"

    STATUS_PENDING = 1
    STATUS_REJECTED = 2
    STATUS_APPROVED = 3
    STATUS_CHOICES = (
        (STATUS_PENDING, _("pending")),
        (STATUS_REJECTED, _("rejected")),
        (STATUS_APPROVED, _("approved")),
    )

    closed = models.DateTimeField(
        blank=True,
        null=True,
    )
    closed_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        models.PROTECT,
        blank=True,
        null=True,
        related_name='access_requests_closed',
    )
    created = models.DateTimeField(auto_now_add=True)
    reason_request = models.TextField(
        blank=True,
        null=True,
    )
    reason_rejected = models.TextField(
        blank=True,
        null=True,
    )
    requester = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        models.PROTECT,
        related_name='access_requests_created',
    )
    reviewers = models.ManyToManyField(
        settings.AUTH_USER_MODEL,
        related_name='access_requests_reviewed',
    )
    secret = models.ForeignKey(
        'Secret',
        models.PROTECT,
        related_name='access_requests',
    )
    status = models.PositiveSmallIntegerField(
        choices=STATUS_CHOICES,
        default=STATUS_PENDING,
    )

    class Meta:
        ordering = ('-created',)

    def __repr__(self):
        return "<AccessRequest {user}@'{name}' ({id})>".format(
            id=self.hashid,
            name=self.secret.name,
            user=self.requester,
        )

    @classmethod
    def get_all_readable_by_user(cls, user):
        if user.is_superuser:
            return cls.objects.all()
        return (
            cls.objects.filter(requester=user) |
            cls.objects.filter(reviewers=user)
        )

    def approve(self, reviewer):
        if self.status != self.STATUS_PENDING:
            raise PermissionDenied(_("Can't approve closed access request"))

        # make sure user is still allowed to handle secret, privileges might
        # have been revoked since the access request was made
        self.secret.check_access(reviewer)

        log(
            _("{reviewer} has approved access request #{access_request} for {requester}, "
              "allowing access to '{secret}'").format(
                access_request=self.id,
                requester=self.requester,
                reviewer=reviewer,
                secret=self.secret.name,
            ),
            actor=reviewer,
            secret=self.secret,
            user=self.requester,
        )

        self.closed = now()
        self.closed_by = reviewer
        self.status = self.STATUS_APPROVED
        self.save()

        self.secret.allowed_users.add(self.requester)

        other_reviewers = list(self.reviewers.all())
        try:
            other_reviewers.remove(reviewer)
        except ValueError:
            # review by other superuser
            pass

        send_mail(
            other_reviewers + [self.requester],
            _("[TeamVault] Access request for '{}' approved").format(self.secret.name),
            "secrets/mail_access_request_approved",
            context={
                'approved_by': reviewer.username,
                'base_url': settings.BASE_URL,
                'secret_name': self.secret.name,
                'secret_url': self.secret.get_absolute_url(),
                'username': self.requester.username,
            },
            user_from=reviewer,
        )

    def assign_reviewers(self):
        candidates = list(self.secret.notify_on_access_request.filter(is_active=True))
        if len(candidates) < 3:
            candidates += list(
                self.secret.allowed_users.order_by('-last_login').filter(is_active=True)[:10]
            )
            for group in self.secret.allowed_groups.all():
                candidates += list(group.user_set.order_by('-last_login').filter(is_active=True)[:3])
        if len(candidates) < 3:
            candidates += list(User.objects.filter(
                is_active=True,
                is_superuser=True,
            ).order_by('-last_login')[:3])
        candidates = set(candidates)
        selected = sample(candidates, min(3, len(candidates)))
        if not selected:
            raise RuntimeError(_("unable to find reviewers for {}").format(self))
        self.reviewers = selected

        send_mail(
            self.reviewers.all(),
            _("[TeamVault] Review access request for '{}'").format(self.secret.name),
            "secrets/mail_access_request_review",
            context={
                'access_request_url': reverse(
                    'secrets.access_request-detail',
                    kwargs={'hashid': self.hashid},
                ),
                'base_url': settings.BASE_URL,
                'secret_name': self.secret.name,
                'secret_url': self.secret.get_absolute_url(),
                'username': self.requester.username,
            },
            user_from=self.requester,
        )

    def reject(self, reviewer, reason=None):
        if self.status != self.STATUS_PENDING:
            raise PermissionDenied(_("Can't reject closed access request"))

        # make sure user is still allowed to handle secret, privileges might
        # have been revoked since the access request was made
        self.secret.check_access(reviewer)

        log(
            _("{reviewer} has rejected access request #{access_request} for {requester}, "
              "NOT allowing access to '{secret}'").format(
                access_request=self.id,
                requester=self.requester,
                reviewer=reviewer,
                secret=self.secret.name,
            ),
            actor=reviewer,
            secret=self.secret,
            user=self.requester,
        )

        self.closed = now()
        self.closed_by = reviewer
        self.reason_rejected = reason
        self.status = self.STATUS_REJECTED
        self.save()

        other_reviewers = list(self.reviewers.all())
        try:
            other_reviewers.remove(reviewer)
        except ValueError:
            # review by other superuser
            pass

        send_mail(
            other_reviewers + [self.requester],
            _("[TeamVault] Access request for '{}' denied").format(self.secret.name),
            "secrets/mail_access_request_denied",
            context={
                'base_url': settings.BASE_URL,
                'denied_by': reviewer.username,
                'reason': reason,
                'secret_name': self.secret.name,
                'secret_url': self.secret.get_absolute_url(),
                'username': self.requester.username,
            },
            user_from=reviewer,
        )

    @property
    def full_url(self):
        return settings.BASE_URL.rstrip("/") + self.get_absolute_url()

    def get_absolute_url(self):
        return reverse('secrets.access_request-detail', args=[str(self.hashid)])

    def is_readable_by_user(self, user):
        return (
            user == self.requester or
            user in self.reviewers.all() or
            user.is_superuser
        )


class Secret(HashIDModel):
    HASHID_NAMESPACE = "Secret"

    ACCESS_POLICY_REQUEST = 1
    ACCESS_POLICY_ANY = 2
    ACCESS_POLICY_HIDDEN = 3
    ACCESS_POLICY_CHOICES = (
        (ACCESS_POLICY_REQUEST, _("request")),
        (ACCESS_POLICY_ANY, _("everyone")),
        (ACCESS_POLICY_HIDDEN, _("hidden")),
    )
    CONTENT_PASSWORD = 1
    CONTENT_CC = 2
    CONTENT_FILE = 3
    CONTENT_CHOICES = (
        (CONTENT_PASSWORD, _("Password")),
        (CONTENT_CC, _("Credit Card")),
        (CONTENT_FILE, _("File")),
    )
    STATUS_OK = 1
    STATUS_NEEDS_CHANGING = 2
    STATUS_DELETED = 3
    STATUS_CHOICES = (
        (STATUS_OK, _("OK")),
        (STATUS_NEEDS_CHANGING, _("needs changing")),
        (STATUS_DELETED, _("deleted")),
    )

    access_policy = models.PositiveSmallIntegerField(
        choices=ACCESS_POLICY_CHOICES,
        default=ACCESS_POLICY_REQUEST,
    )
    allowed_groups = models.ManyToManyField(
        Group,
        blank=True,
        related_name='allowed_passwords',
    )
    allowed_users = models.ManyToManyField(
        settings.AUTH_USER_MODEL,
        blank=True,
        related_name='allowed_passwords',
    )
    content_type = models.PositiveSmallIntegerField(
        choices=CONTENT_CHOICES,
        default=CONTENT_PASSWORD,
    )
    created = models.DateTimeField(auto_now_add=True)
    created_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        models.PROTECT,
        related_name='passwords_created',
    )
    current_revision = models.ForeignKey(
        'SecretRevision',
        models.PROTECT,
        blank=True,
        null=True,
        related_name='_password_current_revision',
    )
    description = models.TextField(
        blank=True,
        null=True,
    )
    filename = models.CharField(
        blank=True,
        max_length=255,
        null=True,
    )
    last_changed = models.DateTimeField(
        auto_now_add=True,
    )
    last_read = models.DateTimeField(
        default=now,
    )
    name = models.CharField(max_length=92)
    needs_changing_on_leave = models.BooleanField(
        default=True,
    )
    notify_on_access_request = models.ManyToManyField(
        settings.AUTH_USER_MODEL,
        blank=True,
        related_name='notify_on_access_requests_for',
    )
    status = models.PositiveSmallIntegerField(
        choices=STATUS_CHOICES,
        default=STATUS_OK,
    )
    url = models.CharField(
        blank=True,
        max_length=255,
        null=True,
        # Django's builtin URL validation is pretty strict to the point
        # of rejecting perfectly good URLs, thus we roll our own very
        # liberal validation
        validators=[validate_url],
    )
    username = models.CharField(
        blank=True,
        max_length=255,
        null=True,
    )

    search_index = SearchVectorField(blank=True, null=True)

    class Meta:
        ordering = ('name', 'username')

    def __str__(self):
        return self.name

    def __repr__(self):
        return "<Secret '{name}' ({id})>".format(id=self.hashid, name=self.name)

    def check_access(self, user):
        if not self.is_visible_to_user(user):
            raise Http404
        elif not self.is_readable_by_user(user):
            raise PermissionDenied()

    @property
    def full_url(self):
        return settings.BASE_URL.rstrip("/") + self.get_absolute_url()

    def get_absolute_url(self):
        return reverse('secrets.secret-detail', args=[str(self.hashid)])

    def get_data(self, user):
        if not self.current_revision:
            raise Http404
        if not self.is_readable_by_user(user):
            log(_(
                    "{user} tried to access '{name}' without permission"
                ).format(
                    name=self.name,
                    user=user.username,
                ),
                actor=user,
                level='warn',
                secret=self,
            )
            raise PermissionError(_(
                "{user} not allowed access to '{name}' ({id})"
            ).format(
                id=self.id,
                name=self.name,
                user=user.username,
            ))
        f = Fernet(settings.TEAMVAULT_SECRET_KEY)
        log(_(
                "{user} read '{name}'"
            ).format(
                name=self.name,
                user=user.username,
            ),
            actor=user,
            level='info',
            secret=self,
            secret_revision=self.current_revision,
        )
        self.current_revision.accessed_by.add(user)
        self.current_revision.save()
        self.last_read = now()
        self.save()

        plaintext_data = f.decrypt(self.current_revision.encrypted_data.tobytes())
        if self.content_type != Secret.CONTENT_FILE:
            plaintext_data = plaintext_data.decode('utf-8')
        return plaintext_data

    @classmethod
    def get_all_readable_by_user(cls, user):
        if user.is_superuser:
            return cls.objects.all()
        return (
            cls.objects.filter(access_policy=cls.ACCESS_POLICY_ANY) |
            cls.objects.filter(allowed_users=user) |
            cls.objects.filter(allowed_groups__in=user.groups.all())
        ).exclude(status=cls.STATUS_DELETED).distinct()

    @classmethod
    def get_all_visible_to_user(cls, user, queryset=None):
        if queryset is None:
            queryset = cls.objects.all()
        if user.is_superuser:
            return queryset
        return (
            queryset.filter(access_policy__in=(cls.ACCESS_POLICY_ANY, cls.ACCESS_POLICY_REQUEST)) |
            queryset.filter(allowed_users=user) |
            queryset.filter(allowed_groups__in=user.groups.all())
        ).exclude(status=cls.STATUS_DELETED).distinct()

    @classmethod
    def get_most_used_for_user(cls, user, limit=5):
        since = now() - timedelta(days=90)
        accessed_secrets = LogEntry.objects.filter(
            actor=user,
            secret__isnull=False,
            time__gte=since,
        ).order_by(
            'secret'
        ).values(
            'secret'
        ).annotate(
            access_count=models.Count('secret'),
        )
        ordered_secrets = sorted(accessed_secrets, key=itemgetter('access_count'), reverse=True)
        return [cls.objects.get(id=item['secret']) for item in ordered_secrets[:limit]]

    @classmethod
    def get_most_recently_used_for_user(cls, user, limit=5):
        secrets = []
        log_entries = LogEntry.objects.filter(
            actor=user,
            secret__isnull=False,
        )
        number_of_log_entries = log_entries.count()
        offset = limit
        while len(secrets) < limit and (offset < number_of_log_entries or offset == limit):
            for log_entry in log_entries[:offset].select_related('secret'):
                if log_entry.secret not in secrets and len(secrets) < limit:
                    secrets.append(log_entry.secret)
            offset += limit
        return secrets

    @classmethod
    def get_search_results(cls, user, term, limit=None):
        base_queryset = cls.get_all_visible_to_user(user)
        name_hits = base_queryset.filter(name__icontains=term)
        fulltext_hits = cls.get_all_visible_to_user(
            user,
            queryset=cls.objects.filter(search_index=term),
        )
        substr_hits = base_queryset.filter(
            models.Q(filename__icontains=term) |
            models.Q(url__icontains=term) |
            models.Q(username__icontains=term)
        )
        if limit:
            name_hits = name_hits[:limit]
            fulltext_hits = fulltext_hits[:limit]
            substr_hits = substr_hits[:limit]
        # concatenate and remove duplicates
        result = (name_hits | fulltext_hits | substr_hits).distinct()
        if limit:
            return result[:limit]
        else:
            return result

    def is_readable_by_user(self, user):
        return (
            user.is_superuser or (
                (
                    self.access_policy == self.ACCESS_POLICY_ANY or
                    user in self.allowed_users.all() or
                    set(self.allowed_groups.all()).intersection(set(user.groups.all()))
                ) and self.status != self.STATUS_DELETED
            )
        )

    def is_visible_to_user(self, user):
        return (
            user.is_superuser or (
                (
                    self.access_policy in (self.ACCESS_POLICY_ANY, self.ACCESS_POLICY_REQUEST) or
                    self.is_readable_by_user(user)
                ) and self.status != self.STATUS_DELETED
            )
       )

    def set_data(self, user, plaintext_data, skip_access_check=False):
        # skip_access_check is used when initially creating a secret
        # and makes it possible to create secrets you don't have access
        # to
        if not skip_access_check and not self.is_readable_by_user(user):
            raise PermissionError(_(
                "{user} not allowed access to '{name}' ({id})"
            ).format(
                id=self.id,
                name=self.name,
                user=user.username,
            ))
        # save the length before encoding so multi-byte characters don't
        # mess up the result
        plaintext_length = len(plaintext_data)
        if isinstance(plaintext_data, str):
            plaintext_data = plaintext_data.encode('utf-8')
        plaintext_data_sha256 = sha256(plaintext_data).hexdigest()
        f = Fernet(settings.TEAMVAULT_SECRET_KEY)
        encrypted_data = f.encrypt(plaintext_data)
        try:
            # see the comment on unique_together for SecretRevision
            p = SecretRevision.objects.get(
                plaintext_data_sha256=plaintext_data_sha256,
                secret=self,
            )
        except SecretRevision.DoesNotExist:
            p = SecretRevision()
        p.encrypted_data = encrypted_data
        p.length = plaintext_length
        # the hash is needed for unique_together (see below)
        # unique_together uses an index on its fields which is
        # problematic with the largeish blobs we might store here (see
        # issue #30)
        p.plaintext_data_sha256 = plaintext_data_sha256
        p.secret = self
        p.set_by = user
        p.save()
        p.accessed_by.add(user)
        if self.current_revision:
            previous_revision_id = self.current_revision.id
        else:
            previous_revision_id = _("none")
        self.current_revision = p
        self.last_changed = now()
        self.last_read = now()
        if self.status == self.STATUS_NEEDS_CHANGING:
            self.status = self.STATUS_OK
        self.save()
        log(_(
                "{user} set a new secret for '{name}' ({oldrev}->{newrev})"
            ).format(
                name=self.name,
                newrev=self.current_revision.id,
                oldrev=previous_revision_id,
                user=user.username,
            ),
            actor=user,
            level='info',
            secret=self,
            secret_revision=self.current_revision,
        )


class SecretRevision(HashIDModel):
    HASHID_NAMESPACE = "SecretRevision"

    accessed_by = models.ManyToManyField(
        settings.AUTH_USER_MODEL,
    )
    created = models.DateTimeField(auto_now_add=True)
    encrypted_data = models.BinaryField()
    length = models.PositiveIntegerField(
        default=0,
    )
    plaintext_data_sha256 = models.CharField(
        max_length=64,
    )
    secret = models.ForeignKey(
        Secret,
        models.PROTECT,
    )
    set_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        models.PROTECT,
        related_name='password_revisions_set',
    )

    class Meta:
        ordering = ('-created',)
        # Consider the following scenario:
        # 1. an employee reads a password, it is "secret123"
        # 2. the password is changed to "secret234"
        # 3. the password is changed back to "secret123"
        # even though the employee never read the third instance of the
        # password, they still know the password from step 1.
        # This unique_together forces passwords 1 and 3 to be the same
        # database object, retaining accessed_by. This way, when the
        # employee leaves, password 3 is correctly assumed known to
        # the employee.
        unique_together = (('plaintext_data_sha256', 'secret'),)

    def __repr__(self):
        return "<SecretRevision '{name}' ({id})>".format(id=self.hashid, name=self.secret.name)

    @property
    def is_current_revision(self):
        return self.secret.current_revision == self


@receiver(post_save, sender=Secret)
def update_search_index(sender, **kwargs):
    Secret.objects.filter(id=kwargs['instance'].id).update(
        search_index=(
            SearchVector('name', weight='A') +
            SearchVector('description', weight='B') +
            SearchVector('username', weight='C') +
            SearchVector('filename', weight='D')
        ),
    )



from django.conf.urls import url

from .api import AccessRequestDetail, AccessRequestList, SecretDetail, SecretList, \
    SecretRevisionDetail, data_get


urlpatterns = (
    url(
        r'^access-requests/$',
        AccessRequestList.as_view(),
        name='api.access-request_list',
    ),
    url(
        r'^access-requests/(?P<hashid>\w+)/$',
        AccessRequestDetail.as_view(),
        name='api.access-request_detail',
    ),
    url(
        r'^secrets/$',
        SecretList.as_view(),
        name='api.secret_list',
    ),
    url(
        r'^secrets/(?P<hashid>\w+)/$',
        SecretDetail.as_view(),
        name='api.secret_detail',
    ),
    url(
        r'^secret-revisions/(?P<hashid>\w+)/$',
        SecretRevisionDetail.as_view(),
        name='api.secret-revision_detail',
    ),
    url(
        r'^secret-revisions/(?P<hashid>\w+)/data$',
        data_get,
        name='api.secret-revision_data',
    ),
)



# Generated by Django 2.0 on 2018-02-20 13:56

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('secrets', '0018_auto_20180220_1244'),
    ]

    operations = [
        migrations.AddField(
            model_name='secret',
            name='notify_on_access_request',
            field=models.ManyToManyField(blank=True, related_name='notify_on_access_requests_for', to=settings.AUTH_USER_MODEL),
        ),
    ]



# Generated by Django 2.0 on 2018-02-20 12:44

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('secrets', '0017_auto_20180220_1115'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='secretrevision',
            unique_together={('plaintext_data_sha256', 'secret')},
        ),
        migrations.RemoveField(
            model_name='secretrevision',
            name='encrypted_data_sha256',
        ),
    ]



from django.contrib.auth.decorators import user_passes_test
from django.contrib.auth.models import User
from django.db.models import Q
from django.shortcuts import get_object_or_404
from django.views.generic import ListView

from ..secrets.models import Secret
from .models import LogEntry


class LogEntryList(ListView):
    context_object_name = 'log_entries'
    paginate_by = 100
    template_name = "audit/log.html"

    def get_context_data(self, **kwargs):
        context = super(LogEntryList, self).get_context_data(**kwargs)
        context['secret'] = None
        context['user'] = None
        if "secret" in self.request.GET:
            context['secret'] = get_object_or_404(Secret, hashid=self.request.GET['secret'])
        elif "user" in self.request.GET:
            context['user'] = get_object_or_404(User, username=self.request.GET['user'])
        return context

    def get_queryset(self):
        if "secret" in self.request.GET:
            secret = get_object_or_404(Secret, hashid=self.request.GET['secret'])
            return LogEntry.objects.filter(secret=secret)
        elif "user" in self.request.GET:
            user = get_object_or_404(User, username=self.request.GET['user'])
            return LogEntry.objects.filter(Q(actor=user) | Q(user=user))
        elif "search" in self.request.GET:
            raise NotImplementedError
        else:
            return LogEntry.objects.all()
auditlog = user_passes_test(lambda u: u.is_superuser)(LogEntryList.as_view())



# Generated by Django 2.2.4 on 2019-08-22 12:34

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('secrets', '0021_auto_20180220_1428'),
    ]

    operations = [
        migrations.AddField(
            model_name='secret',
            name='last_changed',
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now),
            preserve_default=False,
        ),
    ]



from ... import VERSION_STRING
from .models import AccessRequest


def access_request_count(request):
    if request.user.is_authenticated:
        return {
            'access_request_count': AccessRequest.objects.filter(
                reviewers=request.user,
                status=AccessRequest.STATUS_PENDING,
            ).count(),
        }
    else:
        return {}


def version(request):
    return {'version': VERSION_STRING}



from django.conf.urls import include, url

handler404 = 'teamvault.views.handler404'

urlpatterns = (
    url(r'^api/', include('teamvault.apps.secrets.api_urls'), name='api'),
    url(r'^api-auth/', include('rest_framework.urls', namespace='rest_framework')),
    url(r'^audit', include('teamvault.apps.audit.urls'), name='audit'),
    url(r'', include('teamvault.apps.secrets.urls'), name='secrets'),
    url(r'', include('teamvault.apps.accounts.urls'), name='accounts'),
)



# Generated by Django 2.0 on 2018-02-20 11:15

from django.db import migrations


def remove_duplicate_srevs(apps, schema_editor):
    LogEntry = apps.get_model('audit', 'LogEntry')
    Secret = apps.get_model('secrets', 'Secret')
    SecretRevision = apps.get_model('secrets', 'SecretRevision')
    for secret in Secret.objects.all():
        revisions = SecretRevision.objects.filter(secret=secret).order_by('-id')
        hashes_to_revisions = {}
        for revision in revisions:
            if revision.plaintext_data_sha256 in hashes_to_revisions:
                correct_revision = hashes_to_revisions[revision.plaintext_data_sha256]
                assert revision != revision.secret.current_revision
                for log_entry in LogEntry.objects.filter(secret_revision=revision):
                    log_entry.secret_revision = correct_revision
                    log_entry.save()
                correct_revision.accessed_by.add(*list(revision.accessed_by.all()))
                revision.delete()
            else:
                hashes_to_revisions[revision.plaintext_data_sha256] = revision


class Migration(migrations.Migration):

    dependencies = [
        ('secrets', '0016_auto_20180220_1053'),
    ]

    operations = [
        migrations.RunPython(remove_duplicate_srevs),
    ]



# Generated by Django 2.0 on 2018-02-20 10:53
from hashlib import sha256

from cryptography.fernet import Fernet
from django.conf import settings
from django.db import migrations


def backfill_plaintext_hashes(apps, schema_editor):
    SecretRevision = apps.get_model('secrets', 'SecretRevision')
    f = Fernet(settings.TEAMVAULT_SECRET_KEY)
    for srev in SecretRevision.objects.all():
        plaintext_data = f.decrypt(srev.encrypted_data.tobytes())
        srev.plaintext_data_sha256 = sha256(plaintext_data).hexdigest()
        srev.save()


class Migration(migrations.Migration):

    dependencies = [
        ('secrets', '0015_secretrevision_plaintext_data_sha256'),
    ]

    operations = [
        migrations.RunPython(backfill_plaintext_hashes),
    ]



from django.core.mail import EmailMultiAlternatives
from django.template import Context
from django.template.loader import get_template, TemplateDoesNotExist
from django.utils import translation


def send_mail(users_to, subject, template,
              user_from=None, context={}, lang="en",
              attachments=None):
    if attachments is None:
        attachments = []
    c = Context(context)
    translation.activate(lang)
    text_mail = get_template(template + ".txt").render(c)

    msg = EmailMultiAlternatives(
        subject,
        text_mail,
        user_from.email,
        [user.email for user in users_to],
    )

    try:
        html_mail = get_template(template + ".html").render(c)
        msg.attach_alternative(html_mail, "text/html")
    except TemplateDoesNotExist:
        pass

    for filename, data, content_type in attachments:
        msg.attach(filename, data, content_type)

    msg.send()



from os import environ

from setuptools import setup, find_packages

# included here for running tests
environ.setdefault("DJANGO_SETTINGS_MODULE", "teamvault.settings")
environ.setdefault("TEAMVAULT_CONFIG_FILE", "/etc/teamvault.cfg")

setup(
    name="teamvault",
    version="0.8.1",
    description="Keep your passwords behind the firewall",
    author="Torsten Rehn",
    author_email="torsten@rehn.email",
    license="GPLv3",
    url="https://github.com/trehn/teamvault",
    packages=find_packages(),
    include_package_data=True,
    test_suite="tests",
    entry_points={
        'console_scripts': [
            "teamvault=teamvault.cli:main",
        ],
    },
    keywords=["password", "safe", "manager", "sharing"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Framework :: Django",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: Office/Business",
        "Topic :: Security",
    ],
    install_requires=[  # pipenv lock -r | sed "s/ --hash.*//g"
        "asn1crypto==0.24.0",
        "cffi==1.12.3",
        "cryptography==2.7",
        "dj-static==0.0.6",
        "django-auth-ldap==2.0.0",
        "django-filter==2.2.0",
        "django-gravatar2==1.4.2",
        "django==2.2.5",
        "djangorestframework==3.10.3",
        "gunicorn==19.9.0",
        "hashids==1.2.0",
        "psycopg2==2.8.3",
        "pyasn1-modules==0.2.6",
        "pyasn1==0.4.7",
        "pycparser==2.19",
        "python-ldap==3.2.0",
        "pytz==2019.2",
        "six==1.12.0",
        "sqlparse==0.3.0",
        "static3==0.7.0",
    ],
    zip_safe=False,
)



from urllib.parse import urlencode

from django import template


register = template.Library()


@register.simple_tag(takes_context=True)
def replace_get_param(context, **kwargs):
    query = context['request'].GET.dict()
    query.update(kwargs)
    return urlencode(query)


@register.filter
def smart_pages(all_pages, current_page):
    all_pages = list(all_pages)
    smart_pages = set([
        1,
        all_pages[-1],
        current_page,
        max(min(current_page // 2, all_pages[-1]), 1),
        max(min(current_page + ((all_pages[-1] - current_page) // 2), all_pages[-1]), 1),
        max(min(current_page + 1, all_pages[-1]), 1),
        max(min(current_page + 2, all_pages[-1]), 1),
        max(min(current_page - 1, all_pages[-1]), 1),
        max(min(current_page - 2, all_pages[-1]), 1),
    ])
    return sorted(smart_pages)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2016-09-30 10:41
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('secrets', '0009_auto_20150322_0949'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='secret',
            name='search_index',
        ),
    ]



from os.path import dirname, join, realpath

from .apps.settings.config import (
    configure_database,
    configure_django_secret_key,
    configure_hashid,
    configure_logging,
    configure_session,
    get_config,
)

CONFIG = get_config()
PROJECT_ROOT = realpath(dirname(__file__))

### Django

AUTHENTICATION_BACKENDS = (
    'django.contrib.auth.backends.ModelBackend',
)

DATABASES = configure_database(CONFIG)

FILE_UPLOAD_HANDLERS = (
    "teamvault.apps.secrets.utils.CappedMemoryFileUploadHandler",
)

FIXTURE_DIRS = (
    join(PROJECT_ROOT, "fixtures"),
)

INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.humanize',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django_gravatar',
    'rest_framework',
    'teamvault.apps.accounts.AccountsConfig',
    'teamvault.apps.audit.AuditConfig',
    'teamvault.apps.secrets.SecretsConfig',
    'teamvault.apps.settings.SettingsConfig',
]

LANGUAGE_CODE = "en-us"

LOCALE_PATHS = (PROJECT_ROOT + "/locale",)

LOGIN_REDIRECT_URL = "/"
LOGIN_URL = 'accounts.login'
LOGOUT_URL = 'accounts.logout'

LOGGING = configure_logging(CONFIG)

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.locale.LocaleMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = "teamvault.urls"

SECRET_KEY = configure_django_secret_key(CONFIG)

SECURE_PROXY_SSL_HEADER = ("HTTP_X_FORWARDED_PROTO", "https")

SESSION_ENGINE = 'django.contrib.sessions.backends.signed_cookies'
SESSION_COOKIE_AGE, SESSION_EXPIRE_AT_BROWSER_CLOSE, SESSION_COOKIE_SECURE = \
    configure_session(CONFIG)

STATIC_ROOT = join(PROJECT_ROOT, "static_collected")

# remember this is hardcoded in the error page templates (e.g. 500.html)
STATIC_URL = "/static/"

STATICFILES_DIRS = (
    join(PROJECT_ROOT, "static"),
)

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [join(PROJECT_ROOT, "templates"),],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.contrib.auth.context_processors.auth',
                'django.template.context_processors.csrf',
                'django.template.context_processors.debug',
                'django.template.context_processors.i18n',
                'django.template.context_processors.media',
                'django.template.context_processors.request',
                'django.template.context_processors.static',
                'teamvault.apps.secrets.context_processors.access_request_count',
                'teamvault.apps.secrets.context_processors.version',
            ],
        },
    },
]

TEST_RUNNER = 'django.test.runner.DiscoverRunner'

TIME_ZONE = 'UTC'

USE_I18N = False
USE_L10N = True
USE_THOUSAND_SEPARATOR = False
USE_TZ = True

### Hashid

HASHID_MIN_LENGTH, HASHID_SALT = configure_hashid(CONFIG)

### REST Framework

REST_FRAMEWORK = {
    'DEFAULT_MODEL_SERIALIZER_CLASS':
        'rest_framework.serializers.HyperlinkedModelSerializer',
    'DEFAULT_PAGINATION_CLASS': 'rest_framework.pagination.PageNumberPagination',
    'DEFAULT_PERMISSION_CLASSES': (
        'rest_framework.permissions.IsAuthenticated',
    ),
    'PAGE_SIZE': 25,
}



from django.conf.urls import url

from . import views

urlpatterns = (
    url(
        r'^$',
        views.dashboard,
        name='dashboard',
    ),
    url(
        r'^access_requests/$',
        views.access_request_list,
        name='secrets.access_request-list',
    ),
    url(
        r'^access_requests/(?P<hashid>\w+)$',
        views.access_request_detail,
        name='secrets.access_request-detail',
    ),
    url(
        r'^access_requests/(?P<hashid>\w+)/(?P<action>(allow|deny))$',
        views.access_request_review,
        name='secrets.access_request-review',
    ),
    url(
        r'^opensearch.xml$',
        views.opensearch,
        name='opensearch',
    ),
    url(
        r'^secrets/$',
        views.secret_list,
        name='secrets.secret-list',
    ),
    url(
        r'^secrets/(?P<hashid>\w+)$',
        views.secret_detail,
        name='secrets.secret-detail',
    ),
    url(
        r'^secrets/(?P<hashid>\w+)/delete$',
        views.secret_delete,
        name='secrets.secret-delete',
    ),
    url(
        r'^secrets/(?P<hashid>\w+)/download$',
        views.secret_download,
        name='secrets.secret-download',
    ),
    url(
        r'^secrets/(?P<hashid>\w+)/edit$',
        views.secret_edit,
        name='secrets.secret-edit',
    ),
    url(
        r'^secrets/(?P<hashid>\w+)/request_access$',
        views.access_request_create,
        name='secrets.secret-request_access',
    ),
    url(
        r'^secrets/(?P<hashid>\w+)/restore$',
        views.secret_restore,
        name='secrets.secret-restore',
    ),
    url(
        r'^secrets/(?P<hashid>\w+)/share$',
        views.secret_share,
        name='secrets.secret-share',
    ),
    url(
        r'^secrets/add/(?P<content_type>\w+)$',
        views.secret_add,
        name='secrets.secret-add',
    ),
    url(
        r'^secrets/live-search$',
        views.secret_search,
        name='secrets.secret-search',
    ),
)



from random import choice
from string import ascii_letters, digits, punctuation

from django.conf import settings
from django.core.files.uploadhandler import MemoryFileUploadHandler, StopUpload


def generate_password(length=12, alphanum=False):
    """
    Returns a password of the given length.
    """
    char_pool = ascii_letters + digits
    if not alphanum:
        char_pool += punctuation
    return "".join(choice(char_pool) for i in range(length))


class CappedMemoryFileUploadHandler(MemoryFileUploadHandler):
    def receive_data_chunk(self, raw_data, start):
        if start + len(raw_data) > settings.TEAMVAULT_MAX_FILE_SIZE:
            raise StopUpload(connection_reset=True)
        super(CappedMemoryFileUploadHandler, self).receive_data_chunk(raw_data, start)



from json import dumps, loads
from urllib.parse import quote, urlencode

from django.conf import settings
from django.contrib.auth.decorators import login_required, user_passes_test
from django.contrib.auth.models import Group, User
from django.core.exceptions import PermissionDenied
from django.http import HttpResponse, HttpResponseRedirect, Http404
from django.shortcuts import get_object_or_404, render
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _
from django.views.decorators.http import require_http_methods
from django.views.generic import DetailView, ListView, TemplateView
from django.views.generic.edit import CreateView, UpdateView
import django_filters

from ..audit.auditlog import log
from .forms import CCForm, FileForm, PasswordForm
from .models import AccessRequest, Secret

ACCESS_STR_IDS = {
    'ACCESS_POLICY_ANY': str(Secret.ACCESS_POLICY_ANY),
    'ACCESS_POLICY_REQUEST': str(Secret.ACCESS_POLICY_REQUEST),
    'ACCESS_POLICY_HIDDEN': str(Secret.ACCESS_POLICY_HIDDEN),
}
CONTENT_TYPE_FORMS = {
    'cc': CCForm,
    'file': FileForm,
    'password': PasswordForm,
}
CONTENT_TYPE_IDS = {
    'cc': Secret.CONTENT_CC,
    'file': Secret.CONTENT_FILE,
    'password': Secret.CONTENT_PASSWORD,
}
CONTENT_TYPE_IDENTIFIERS = {v: k for k, v in CONTENT_TYPE_IDS.items()}
_CONTENT_TYPES = dict(Secret.CONTENT_CHOICES)
CONTENT_TYPE_NAMES = {
    'cc': _CONTENT_TYPES[Secret.CONTENT_CC],
    'file': _CONTENT_TYPES[Secret.CONTENT_FILE],
    'password': _CONTENT_TYPES[Secret.CONTENT_PASSWORD],
}


def _patch_post_data(POST, fields):
    """
    Select2 passes in selected values as CSV instead of as a real
    multiple value field, so we need to split them before any validation
    takes place.
    """
    POST = POST.copy()
    for csv_field in fields:
        if POST.getlist(csv_field) == ['']:
            del POST[csv_field]
        else:
            POST.setlist(
                csv_field,
                POST.getlist(csv_field)[0].split(","),
            )
    return POST


@login_required
def access_request_create(request, hashid):
    secret = Secret.objects.get(hashid=hashid)
    if not secret.is_visible_to_user(request.user):
        raise Http404
    try:
        AccessRequest.objects.get(
            requester=request.user,
            secret=secret,
            status=AccessRequest.STATUS_PENDING,
        )
    except AccessRequest.DoesNotExist:
        if request.method == 'POST' and not secret.is_readable_by_user(request.user):
            access_request = AccessRequest()
            access_request.reason_request = request.POST['reason']
            access_request.requester = request.user
            access_request.secret = secret
            access_request.save()
            access_request.assign_reviewers()
    return HttpResponseRedirect(secret.get_absolute_url())


@login_required
@require_http_methods(["POST"])
def access_request_review(request, hashid, action):
    access_request = get_object_or_404(
        AccessRequest,
        hashid=hashid,
        status=AccessRequest.STATUS_PENDING,
    )
    if not request.user.is_superuser and request.user not in access_request.reviewers.all():
        raise PermissionDenied()

    if action == 'allow':
        access_request.approve(request.user)
    else:
        access_request.reject(request.user, reason=request.POST.get('reason', None))

    return HttpResponseRedirect(reverse('secrets.access_request-list'))


class AccessRequestDetail(DetailView):
    context_object_name = 'access_request'
    model = AccessRequest
    slug_field = 'hashid'
    slug_url_kwarg = 'hashid'
    template_name = "secrets/accessrequest_detail.html"

    def get_object(self):
        if self.request.user.is_superuser:
            return get_object_or_404(
                AccessRequest,
                hashid=self.kwargs['hashid'],
                status=AccessRequest.STATUS_PENDING,
            )
        else:
            return get_object_or_404(
                AccessRequest,
                hashid=self.kwargs['hashid'],
                reviewers=self.request.user,
                status=AccessRequest.STATUS_PENDING,
            )
access_request_detail = login_required(AccessRequestDetail.as_view())


class AccessRequestList(ListView):
    context_object_name = 'access_requests'
    template_name = "secrets/accessrequests_list.html"

    def get_context_data(self, **kwargs):
        queryset = self.get_queryset()
        context = super(AccessRequestList, self).get_context_data(**kwargs)
        context['reviewable'] = queryset.exclude(requester=self.request.user)
        context['pending_review'] = queryset.filter(requester=self.request.user)
        return context

    def get_queryset(self):
        queryset = AccessRequest.get_all_readable_by_user(self.request.user)
        return queryset.filter(status=AccessRequest.STATUS_PENDING)
access_request_list = login_required(AccessRequestList.as_view())


class Dashboard(TemplateView):
    template_name = "secrets/dashboard.html"

    def get_context_data(self, **kwargs):
        context = super(Dashboard, self).get_context_data(**kwargs)
        context['search_term'] = ""
        context['most_used_secrets'] = Secret.get_most_used_for_user(self.request.user)
        context['readable_secrets'] = Secret.get_all_readable_by_user(self.request.user)
        context['recently_used_secrets'] = Secret.get_most_recently_used_for_user(self.request.user)
        return context
dashboard = login_required(Dashboard.as_view())


class OpenSearch(TemplateView):
    content_type = "application/xml"
    template_name = "opensearch.xml"

    def get_context_data(self, **kwargs):
        context = super(OpenSearch, self).get_context_data(**kwargs)
        context['base_url'] = settings.BASE_URL
        return context
opensearch = OpenSearch.as_view()


class SecretAdd(CreateView):
    slug_field = 'hashid'
    slug_url_kwarg = 'hashid'

    def form_valid(self, form):
        secret = Secret()
        secret.content_type = CONTENT_TYPE_IDS[self.kwargs['content_type']]
        secret.created_by = self.request.user

        for attr in ('access_policy', 'description', 'name', 'needs_changing_on_leave', 'url',
                     'username'):
            if attr in form.cleaned_data:
                setattr(secret, attr, form.cleaned_data[attr])
        secret.save()

        for attr in ('notify_on_access_request',):
            if form.cleaned_data[attr]:
                getattr(secret, attr).add(self.request.user)
            else:
                getattr(secret, attr).remove(self.request.user)

        for attr in ('allowed_groups', 'allowed_users'):
            getattr(secret, attr).set(form.cleaned_data[attr])

        if secret.content_type == Secret.CONTENT_PASSWORD:
            plaintext_data = form.cleaned_data['password']
        elif secret.content_type == Secret.CONTENT_FILE:
            plaintext_data = form.cleaned_data['file'].read()
            secret.filename = form.cleaned_data['file'].name
            secret.save()
        elif secret.content_type == Secret.CONTENT_CC:
            plaintext_data = dumps({
                'holder': form.cleaned_data['holder'],
                'number': form.cleaned_data['number'],
                'expiration_month': str(form.cleaned_data['expiration_month']),
                'expiration_year': str(form.cleaned_data['expiration_year']),
                'security_code': str(form.cleaned_data['security_code']),
                'password': form.cleaned_data['password'],
            })
        secret.set_data(self.request.user, plaintext_data, skip_access_check=True)

        return HttpResponseRedirect(secret.get_absolute_url())

    def get_context_data(self, **kwargs):
        context = super(SecretAdd, self).get_context_data(**kwargs)
        try:
            context['pretty_content_type'] = CONTENT_TYPE_NAMES[self.kwargs['content_type']]
        except KeyError:
            raise Http404
        context.update(ACCESS_STR_IDS)
        return context

    def get_form_class(self):
        return CONTENT_TYPE_FORMS[self.kwargs['content_type']]

    def get_template_names(self):
        return "secrets/secret_addedit_{}.html".format(self.kwargs['content_type'])

    def post(self, request, *args, **kwargs):
        request.POST = _patch_post_data(
            request.POST,
            (
                'allowed_groups',
                'allowed_users',
            ),
        )
        return super(SecretAdd, self).post(request, *args, **kwargs)
secret_add = login_required(SecretAdd.as_view())


class SecretEdit(UpdateView):
    context_object_name = 'secret'
    slug_field = 'hashid'
    slug_url_kwarg = 'hashid'

    def form_valid(self, form):
        secret = self.object

        for attr in ('access_policy', 'description', 'name', 'needs_changing_on_leave', 'url',
                     'username'):
            if attr in form.cleaned_data:
                setattr(secret, attr, form.cleaned_data[attr])
        secret.save()

        for attr in ('notify_on_access_request',):
            if form.cleaned_data[attr]:
                getattr(secret, attr).add(self.request.user)
            else:
                getattr(secret, attr).remove(self.request.user)

        for attr in ('allowed_groups', 'allowed_users'):
            getattr(secret, attr).set(form.cleaned_data[attr])

        if secret.content_type == Secret.CONTENT_PASSWORD and form.cleaned_data['password']:
            plaintext_data = form.cleaned_data['password']
        elif secret.content_type == Secret.CONTENT_FILE and form.cleaned_data['file']:
            plaintext_data = form.cleaned_data['file'].read()
            secret.filename = form.cleaned_data['file'].name
            secret.save()
        elif secret.content_type == Secret.CONTENT_CC:
            plaintext_data = dumps({
                'holder': form.cleaned_data['holder'],
                'number': form.cleaned_data['number'],
                'expiration_month': form.cleaned_data['expiration_month'],
                'expiration_year': form.cleaned_data['expiration_year'],
                'security_code': form.cleaned_data['security_code'],
                'password': form.cleaned_data['password'],
            })
        else:
            plaintext_data = None

        if plaintext_data is not None:
            secret.set_data(self.request.user, plaintext_data)

        return HttpResponseRedirect(secret.get_absolute_url())

    def get_context_data(self, **kwargs):
        context = super(SecretEdit, self).get_context_data(**kwargs)
        context['pretty_content_type'] = self.object.get_content_type_display()
        context.update(ACCESS_STR_IDS)
        return context

    def get_form_class(self):
        return CONTENT_TYPE_FORMS[CONTENT_TYPE_IDENTIFIERS[self.object.content_type]]

    def get_initial(self):
        if self.object.content_type == Secret.CONTENT_CC:
            data = loads(self.object.get_data(self.request.user))
            return {
                'holder': data['holder'],
                'number': data['number'],
                'expiration_month': data['expiration_month'],
                'expiration_year': data['expiration_year'],
                'security_code': data['security_code'],
                'password': data['password'],
            }
        else:
            return {}

    def get_object(self, queryset=None):
        secret = get_object_or_404(Secret, hashid=self.kwargs['hashid'])
        secret.check_access(self.request.user)
        return secret

    def get_template_names(self):
        return "secrets/secret_addedit_{}.html".format(CONTENT_TYPE_IDENTIFIERS[self.object.content_type])

    def post(self, request, *args, **kwargs):
        request.POST = _patch_post_data(
            request.POST,
            (
                'allowed_groups',
                'allowed_users',
            ),
        )
        return super(SecretEdit, self).post(request, *args, **kwargs)
secret_edit = login_required(SecretEdit.as_view())


@login_required
def secret_delete(request, hashid):
    secret = get_object_or_404(Secret, hashid=hashid)
    secret.check_access(request.user)
    if request.method == 'POST':
        log(_(
                "{user} deleted '{name}' ({id}:{revision})"
            ).format(
                id=secret.id,
                name=secret.name,
                revision=secret.current_revision.id,
                user=request.user.username,
            ),
            actor=request.user,
            level='info',
            secret=secret,
            secret_revision=secret.current_revision,
        )
        secret.status = Secret.STATUS_DELETED
        secret.save()
        return HttpResponseRedirect(reverse('secrets.secret-list') + "?" + urlencode([("search", secret.name.encode('utf-8'))]))
    else:
        return render(request, "secrets/secret_delete.html", {'secret': secret})


@login_required
@user_passes_test(lambda u: u.is_superuser)
def secret_restore(request, hashid):
    secret = get_object_or_404(Secret, hashid=hashid)
    secret.check_access(request.user)
    if request.method == 'POST':
        log(_(
                "{user} restore '{name}' ({id}:{revision})"
            ).format(
                id=secret.id,
                name=secret.name,
                revision=secret.current_revision.id,
                user=request.user.username,
            ),
            actor=request.user,
            level='info',
            secret=secret,
            secret_revision=secret.current_revision,
        )
        secret.status = Secret.STATUS_OK
        secret.save()
        return HttpResponseRedirect(reverse('secrets.secret-list') + "?" + urlencode([("search", secret.name.encode('utf-8'))]))
    else:
        return render(request, "secrets/secret_restore.html", {'secret': secret})


@login_required
@require_http_methods(["GET"])
def secret_download(request, hashid):
    secret = get_object_or_404(Secret, hashid=hashid)
    if secret.content_type != Secret.CONTENT_FILE:
        raise Http404
    secret.check_access(request.user)

    response = HttpResponse(secret.get_data(request.user))
    response['Content-Disposition'] = \
        "attachment; filename*=UTF-8''{}".format(quote(secret.filename))
    response['Content-Type'] = "application/force-download"
    return response


class SecretDetail(DetailView):
    context_object_name = 'secret'
    model = Secret
    slug_field = 'hashid'
    slug_url_kwarg = 'hashid'
    template_name = "secrets/secret_detail.html"

    def get_context_data(self, **kwargs):
        context = super(SecretDetail, self).get_context_data(**kwargs)
        secret = self.get_object()
        context['content_type'] = CONTENT_TYPE_IDENTIFIERS[secret.content_type]
        context['readable'] = secret.is_readable_by_user(self.request.user)
        context['secret_url'] = reverse(
            'api.secret-revision_data',
            kwargs={'hashid': secret.current_revision.hashid},
        )
        if context['readable']:
            context['placeholder'] = secret.current_revision.length * "•"
        else:
            try:
                context['access_request'] = AccessRequest.objects.get(
                    secret=secret,
                    status=AccessRequest.STATUS_PENDING,
                    requester=self.request.user,
                )
            except AccessRequest.DoesNotExist:
                context['access_request'] = None
        return context

    def get_object(self):
        object = super(SecretDetail, self).get_object()
        if not object.is_visible_to_user(self.request.user):
            raise Http404
        return object
secret_detail = login_required(SecretDetail.as_view())


class SecretFilter(django_filters.FilterSet):
    order = django_filters.OrderingFilter(
        choices=(
            ('last_changed', _("Last changed")),
            ('last_read', _("Last read")),
        ),
    )

    class Meta:
        model = Secret
        fields = {
            'last_changed': ['gt', 'gte', 'lt', 'lte'],
            'last_read': ['gt', 'gte', 'lt', 'lte'],
        }


class SecretList(ListView):
    context_object_name = 'secrets'
    paginate_by = 25
    template_name = "secrets/secret_list.html"

    def get_context_data(self, **kwargs):
        context = super(SecretList, self).get_context_data(**kwargs)
        context['readable_secrets'] = Secret.get_all_readable_by_user(self.request.user)
        return context

    def get_queryset(self):
        if "search" in self.request.GET:
            queryset = Secret.get_search_results(self.request.user, self.request.GET['search'])
        else:
            queryset = Secret.get_all_visible_to_user(self.request.user)
        filtered = SecretFilter(self.request.GET, queryset)
        return filtered.qs
secret_list = login_required(SecretList.as_view())


@login_required
@require_http_methods(["POST"])
def secret_share(request, hashid):
    secret = get_object_or_404(Secret, hashid=hashid)
    secret.check_access(request.user)

    request.POST = _patch_post_data(request.POST, ('share_groups', 'share_users'))

    groups = []
    for group_id in request.POST.getlist('share_groups', []):
        groups.append(get_object_or_404(Group, pk=int(group_id)))

    users = []
    for user_id in request.POST.getlist('share_users', []):
        users.append(get_object_or_404(User, pk=int(user_id)))

    for group in groups:
        log(
            _("{actor} shared '{secret}' with {group}").format(
                actor=request.user,
                group=group.name,
                secret=secret.name,
            ),
            actor=request.user,
            group=group,
            secret=secret,
        )
        secret.allowed_groups.add(group)
        # TODO email with additional message field

    for user in users:
        log(
            _("{actor} shared '{secret}' with {user}").format(
                actor=request.user,
                secret=secret.name,
                user=user.username,
            ),
            actor=request.user,
            secret=secret,
            user=user,
        )
        secret.allowed_users.add(user)
        # TODO email with additional message field

    return HttpResponseRedirect(secret.get_absolute_url())


@login_required
@require_http_methods(["GET"])
def secret_search(request):
    search_term = request.GET['q']
    search_result = []
    filtered_secrets = list(Secret.get_search_results(request.user, search_term, limit=10))
    unreadable_secrets = filtered_secrets[:]
    sorted_secrets = []

    # sort readable passwords to top...
    for secret in filtered_secrets:
        if secret.is_readable_by_user(request.user):
            sorted_secrets.append((secret, "unlock"))
            unreadable_secrets.remove(secret)

    # and others to the bottom
    for secret in unreadable_secrets:
        sorted_secrets.append((secret, "lock"))

    for secret, icon in sorted_secrets:
        search_result.append({
            'name': secret.name,
            'url': reverse('secrets.secret-detail', kwargs={'hashid': secret.hashid}),
            'icon': icon,
        })

    return HttpResponse(dumps(search_result), content_type="application/json")
